# -------------------------------------------------------------------------
#
#        D E S T I N A T I O N   S Y S T E M
#
# -------------------------------------------------------------------------
import abc
import pathlib

from .context import DeliveryContext
from .products import (
    OUS,
    Calibration,
    ExecutionBlock,
    Image,
    ProcessingRequest,
    SpooledProduct,
)


class Destination(abc.ABC):
    """
    Destinations are locations that files can be copied into. They might not
    always be on a local disk; FTP or Globus could also be destinations.

    The destination API is very simply, consisting just of adding files.
    """

    @abc.abstractmethod
    def add_file(self, relative_path: str, file: pathlib.Path):
        pass

    @abc.abstractmethod
    def close(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        # ensure that if we are used as a context manager ('with' statement)
        # that the destinations do get properly closed
        self.close()


class DestinationDecorator(Destination):
    def __init__(self, underlying: Destination):
        self.underlying = underlying

    def add_file(self, relative_path: str, file: pathlib.Path):
        self.underlying.add_file(relative_path, file)

    def close(self):
        self.underlying.close()


class TarDecorator(DestinationDecorator):
    """
    This decorator creates a local tar archive. Calls to add_file
    are intercepted and instead the file contents are added to the
    tar archive. When close() is called, we finalize the tar file
    and place it in the delivery area.
    """

    def __init__(self, context: DeliveryContext, underlying: Destination):
        super().__init__(underlying)
        self.context = context


class ChecksumDecorator(DestinationDecorator):
    """
    This decorator ensures that an MD5SUM file appears in the underlying
    destination after all the files are added, and the contents of that
    file are, as one would expect, the MD5SUMs of the files added to the
    destination, in the format that ``md5sum -c`` expects.
    """

    def __init__(self, context: DeliveryContext, underlying: Destination):
        super().__init__(underlying)
        self.context = context


class SubdirectoryDecorator(DestinationDecorator):
    def __init__(self, underlying: Destination, subdirectory: str):
        super().__init__(underlying)
        self.subdirectory = subdirectory

    def add_file(self, relative_path: str, file: pathlib.Path):
        self.underlying.add_file(self.subdirectory + "/" + relative_path, file)


class LocalDestination(Destination):
    """
    LocalDestination is for delivering to a local directory on the filesystem.
    """

    def __init__(self, context: DeliveryContext, path: pathlib.Path):
        self.context = context
        self.path = path

    def add_file(self, relative_path: str, file: pathlib.Path):
        raise NotImplementedError

    def close(self):
        """
        Nothing special actually needs to be done for local deliveries
        when we close the destination.
        """
        pass


class DestinationBuilder:
    """
    To facilitate building a stack of destination and its decorators.
    """

    def __init__(self, context: DeliveryContext):
        self.context = context
        self._destination = None

    def local(self, path: pathlib.Path):
        """Add a local destination with the given path"""
        self._destination = LocalDestination(self.context, path)
        return self

    def tar(self):
        """Add the tar decorator to the destination"""
        self._destination = TarDecorator(self.context, self._destination)
        return self

    def build(self):
        """Create the destination"""
        # We *always* want checksums, so that goes on last here
        return ChecksumDecorator(self.context, self._destination)
