import pathlib

from .context import DeliveryContext
from .deliverer import ProductDeliverer
from .finder import HeuristicProductFinder, PiperesultsProductFinder


class Delivery:
    def has_piperesults(self, path: pathlib.Path) -> bool:
        raise NotImplementedError

    def create_product_finder(self, source: pathlib.Path) -> ProductFinder:
        """
        Based on the contents of the source/ folder, make the right flavor
        of product finder
        :param source: directory to examine
        :return: a product finder
        """
        # if there is a piperesults file, use that finder
        if self.has_piperesults(source):
            return PiperesultsProductFinder(source)
        else:
            return HeuristicProductFinder(source)

    def deliver(self, context: DeliveryContext):
        """
        Primary entrypoint to delivery process
        :param context:  Settings to configure delivery
        :return:  some sort of user-friendly representation of where everything wound up
        """
        # make the destination
        destination = context.create_destination()

        # find the products
        finder = self.create_product_finder(context.source)

        # loop over the products and deliver them
        for product in finder.find_products():
            product.deliver_to(destination)

        # the last thing we should do is return the URL to the delivered stuff.
        # for a local delivery, this will be a file URL; otherwise it will be an http URL.
        # FIXME: not implemented
        raise NotImplementedError


def main(args=None):
    """CLI entry point"""
    context = DeliveryContext.parse_commandline(args)
    print(context)
    Delivery().deliver(context)
