import abc
import pathlib

from .deliverer import ProductDeliverer, Destination, SubdirectoryDecorator


class SpooledProduct(abc.ABC):
    """
    A SpooledProduct is something interesting enough to deliver to an end-user.
    SpooledProducts might be science products or auxilliary products, but they
    might just be some temporary thing that came out of some processing, never
    to be seen again.

    SpooledProducts have a type and a path. More specific SpooledProducts may
    have other specific metadata unto themselves.
    """

    @abc.abstractmethod
    def deliver_to(self, destination: Destination):
        """
        Deliver sets up a double dispatch so we can get the product type into the
        method itself, and behave differently depending on the type of product
        being delivered.

        :param deliverer: the deliverer to deliver to
        """
        pass


class ExecutionBlock(SpooledProduct):
    """
    An execution block is a collection of SDM and BDF files. There is also metadata
    about the execution block that may be interesting for the purposes of delivery
    eventually.
    """

    def __init__(self, path: pathlib.Path):
        self.path = path

    def deliver_to(self, destination: Destination):
        destination.add_directory(self.path)


# Future types of product that might be needed:
#
# class Calibration(SpooledProduct):
#     def deliver_to(self, destination: Destination):
#         raise NotImplementedError
#
#
# class Image(SpooledProduct):
#     def deliver_to(self, destination: Destination):
#         raise NotImplementedError
#
#
# class ProcessingRequest(SpooledProduct):
#     def deliver_to(self, destination: Destination):
#         raise NotImplementedError
#
#
# class OUS(SpooledProduct):
#     """
#     OUSes have a great deal of sub-products within them.
#     """
#
#     def __init__(self, id):
#         self.id = id
#         self._products = []
#
#     def add_product(self, product):
#         """
#         Add a product to this OUS.
#
#         :param product:  the product to add
#         """
#         self._products.append(product)
#
#     def deliver_to(self, destination: Destination):
#         # FIXME: example implementation; refine
#         # The trick here is to basically "cd" into the OUS directory and then
#         # proceed as normal. The subdirectory decorator will ensure that the
#         # products have the same structure, but inside the OUS directory
#         # instead of the level above.
#         ous_destination = SubdirectoryDecorator(destination, self.id)
#         for product in self._products:
#             product.deliver_to(ous_destination)
#
