"""
Services specific to the Capability System of Workspaces, to be used by our APIs and client programs
"""
import threading
import time
from collections import namedtuple
from queue import Empty, PriorityQueue, Queue
from typing import List, NamedTuple, Optional
from channels.amqp_helpers import Channel, WorkflowEventChannel, CapabilityEventChannel
from sqlalchemy.orm import Session
from .capability_interfaces import (
    CapabilityEngineIF,
    CapabilityExecutionIF,
    CapabilityInfoIF,
    CapabilityName,
    CapabilityQueueIF,
    CapabilityRequestIF,
    CapabilityServiceIF,
    ParameterIF,
)
from .helpers import (
    CapabilityEventType,
    CapabilityRequestState,
    CapabilitySequence,
    CapabilityStepType,
    ExecutionPriority,
    ExecutionState,
    FutureProduct,
    Parameter,
)
from .product_interfaces import FutureProductIF
from .schema import (
    AbstractFile,
    Base,
    Capability,
    CapabilityEvent,
    CapabilityExecution,
    CapabilityRequest,
    CapabilityVersion,
    WorkflowEvent,
    WorkflowEventType,
    WorkflowRequest,
)
from .workflow_interfaces import WorkflowServiceIF


class CapabilityService(CapabilityServiceIF):
    """
    The capability service: clients access this to request capability runs
    """

    def __init__(
            self, capability_info: CapabilityInfoIF, workflow_service: WorkflowServiceIF
    ):
        self.execution_pool = []
        self.queues = {}
        self.capability_info = capability_info
        self.workflow_service = workflow_service
        self.listener = threading.Thread(target=self.listen_for_events, daemon=True)
        self.listener.start()

    def create_request(
            self,
            capability_name: str,
            parameters: List[Parameter] = None,
            products: List[FutureProduct] = None,
    ) -> CapabilityRequestIF:
        """
        Create a capability request for a given capability

        :param capability_name:
        :param parameters: Parameters for request
        :param products: Products the request will create upon success
        :return: Capability request
        """
        return self.capability_info.create_capability_request(
            capability_name, parameters, products
        )

    def run_capability(self, request: CapabilityRequest) -> CapabilityExecutionIF:
        """
        Run a capability with the requested settings

        :param request: Capability request
        :return: Execution of the request
        """
        execution = self.capability_info.create_execution(request)
        steps = CapabilitySequence.from_str(execution.steps)

        if steps[0].step_type == CapabilityStepType.PrepareAndRunWorkflow:
            # First step is to run workflow, move to capability queue
            # FIXME: Priority needs to be dynamic
            self.enqueue_execution(execution, ExecutionPriority.Default.value)
        else:
            self.execution_pool.append(execution)
        return execution

    def update_execution(self, event: WorkflowEvent):
        """
        Update capability execution given a received event
        TODO: Implement execution cancellation

        :param event: Incoming event
        """
        capability_event = self.workflow_to_capability_event(event)

        if not capability_event:
            # WorkflowEvent does not need to update the execution
            return

        execution = self.capability_info.lookup_execution(capability_event.execution_id)
        step_sequence = CapabilitySequence.from_str(execution.steps)
        current_step = step_sequence[execution.current_step]
        event_type = CapabilityEventType[capability_event.event_type]

        # Check to make sure event type is correct
        if current_step.step_type.value == event_type.value:
            # Check if previous step (that just completed) is run workflow step
            if (
                    step_sequence[execution.current_step].step_type
                    == CapabilityStepType.PrepareAndRunWorkflow.name
            ):
                # Return capability engine to available state
                self.queues[execution.capability.id].complete_execution(
                    capability_event.execution_id
                )

            if execution.current_step != len(step_sequence):
                # Execution is not on its last step
                execution.current_step += 1
            else:
                # Execution on its last step and received the proper event
                execution.state = ExecutionState.Complete.name
                self.capability_info.save_execution(execution)
                return

            # Check if upcoming step is run workflow step
            if (
                    step_sequence[execution.current_step].step_type
                    == CapabilityStepType.PrepareAndRunWorkflow
            ):
                # Update execution record's state
                execution.state = ExecutionState.ExecutingStep.name
                # Enqueue execution that is on a run workflow step
                # FIXME: Priority needs to be dynamic; perhaps a priority column in execution schema?
                self.enqueue_execution(execution, ExecutionPriority.Default.value)

            # Update execution record's state
            execution.state = ExecutionState.Ready.name
            self.capability_info.save_execution(execution)
        else:
            # TODO: Logging
            print(
                f"Mismatched event type {event_type} for execution {capability_event.execution_id}"
            )

    def enqueue_execution(
            self,
            execution: CapabilityExecutionIF,
            priority: int = ExecutionPriority.Default.value,
    ):
        """
        Move execution record that is ready to execute a workflow into the appropriate capability
        queue

        :param execution: Capability execution record
        :param priority: Priority of the requested execution
        """
        # Get correct queue or initialize one
        queue = self.queues.get(
            execution.capability.id,
            CapabilityQueue(
                self.capability_info,
                self.workflow_service,
                execution.capability.max_jobs,
            ),
        )
        queue.enqueue(execution, priority)
        self.queues[execution.capability.id] = queue

        # Remove execution record from pool
        if execution in self.execution_pool:
            self.execution_pool.remove(execution)

    def workflow_to_capability_event(self, event: WorkflowEvent) -> CapabilityEvent:
        """
        Convert a workflow event into a capability event

        :param event: Workflow event
        :return: Capability event translated from given workflow event
        """
        if event.type == WorkflowEventType.SUBMITTED.name:
            # Workflow submitted
            event_type = CapabilityEventType.WorkflowReady
        elif event.type == WorkflowEventType.TERMINATED.name and event.retval == 0:
            # Workflow completed successfully
            event_type = CapabilityEventType.WorkflowComplete
        else:
            # No corresponding capability event
            return None

        execution = self.capability_info.lookup_execution_by_workflow_request_id(
            event.workflow_request_id
        )

        return CapabilityEvent(
            event_type, execution.capability_request.id, execution.id
        )

    def listen_for_events(self):
        """
        Method that will run in a different thread that will be listening for capability events that it can use
        to update capability executions
        """
        thread_workflow_events = Channel(WorkflowEventChannel(), threaded=True)
        thread_workflow_events.listen(callback=self.update_execution)


class CapabilityInfo(CapabilityInfoIF):
    """
    Interface for data access object that can look up and record information about capabilities
    and capability requests. Accesses and modifies the database.
    """

    def __init__(self, session: Session):
        self.session = session

    def create_capability(
            self, name: CapabilityName, steps: CapabilitySequence, max_jobs: int
    ) -> Capability:
        """
        Create new capability and save it in the database

        :param name: Name of new capability
        :param steps: List of capability steps required to execute the capability
        :param max_jobs: Max allowed number of concurrent executions of this kind of capability
        :return: Created capability
        """
        capability = Capability(name=name, steps=str(steps), max_jobs=max_jobs)
        self.save_entity(capability)
        return capability

    def create_capability_request(
            self,
            capability_name: str,
            parameters: List[ParameterIF] = None,
            future_products: List[FutureProductIF] = None,
    ) -> CapabilityRequest:
        """
        Create new capability request and save it in the database

        :param capability_name: name of the requested capability
        :param parameters: List
        :param future_products:
        :return: Created CapabilityRequest
        """
        capability = self.lookup_capability(capability_name)
        request = CapabilityRequest(
            state=CapabilityRequestState.Ready.name,
            capability=capability,
            parameters=str(parameters),
            # a trick here is to ensure that we always have a first version, with the original parameters
            versions=[CapabilityVersion(version_number=1, parameters=str(parameters))],
        )
        self.save_entity(request)
        return request

    def create_execution(self, request: CapabilityRequest) -> CapabilityExecution:
        """
        Create new execution record for a request and save it in the database

        :param request: ID of the capability request
        :return: Created CapabilityExecution
        """
        # look up the most recent version
        most_recent_version = request.versions[-1]

        execution = CapabilityExecution(
            state=ExecutionState.Ready.name,
            version=most_recent_version,
            current_step=0,
            # ensure that we have a copy of the step sequence as it was when the execution started
            steps=request.capability.steps,
        )
        self.save_entity(execution)
        return execution

    def lookup_capability(self, capability_name: str) -> Capability:
        return self.session.query(Capability).filter_by(name=capability_name).first()

    def lookup_capability_request(self, request_id) -> CapabilityRequest:
        return self.session.query(CapabilityRequest).filter_by(id=request_id).first()

    def lookup_execution(self, execution_id: int) -> CapabilityExecution:
        return (
            self.session.query(CapabilityExecution).filter_by(id=execution_id).first()
        )

    def save_entity(self, entity: Base):
        """
        Save entity to database

        :param entity:
        """
        self.session.add(entity)
        self.session.flush()

    def lookup_execution_by_workflow_request_id(
            self, workflow_request_id: int
    ) -> CapabilityExecution:
        """
        Find execution record given a capability request

        :param workflow_request_id:
        :return: Capability execution record
        """
        return (
            self.session.query(CapabilityExecution)
                .filter_by(workflow_request_id=workflow_request_id)
                .one()
        )

    def requests_for_capability(self, capability_name: str) -> List[CapabilityRequest]:
        """
        Gets all capability requests for a given capability

        :param capability_name: Name of capability
        :return: List of associated requests
        """
        capability = self.lookup_capability(capability_name)
        return (
            self.session.query(CapabilityRequest)
                .filter_by(capability_id=capability.id)
                .all()
        )

    def save_execution(self, execution: CapabilityExecutionIF):
        self.session.add(execution)


class CapabilityEngine(CapabilityEngineIF):
    """
    Executes a prepare and run workflow step of a capability
    """

    def __init__(
            self,
            capability_info: CapabilityInfoIF,
            workflow_service: "WorkflowServiceIF",
    ):
        self.capability_info = capability_info
        self.workflow_service = workflow_service

    def execute(self, execution: CapabilityExecutionIF):
        """
        Communicate with workflow service and send a request to run a workflow with given settings

        :param execution:
        """
        step_sequence = CapabilitySequence.from_str(execution.steps)
        # Grab value of current step (workflow name)
        cur_step = step_sequence[int(execution.current_step)]
        cur_step.execute(self, execution)

    def submit_workflow_request(
            self, workflow_name: str, workflow_args: dict, files: List[AbstractFile]
    ):
        """
        Submit a workflow request to the workflow service

        :param workflow_name:
        :param workflow_args:
        :param files:
        """
        workflow_request = WorkflowRequest(
            workflow_name=workflow_name, argument=workflow_args
        )
        self.workflow_service.execute(workflow_request, files)


class CapabilityQueue(CapabilityQueueIF):
    """
    Organizes capability executions in a priority order and makes it possible to control
    the number of concurrent executions or pause execution altogether
    """

    EngineList = namedtuple("EngineList", ["available", "in_use"])

    def __init__(
            self,
            capability_info: CapabilityInfoIF,
            workflow_service: WorkflowServiceIF,
            max_concurrency: int,
    ):
        self.capability_info = capability_info
        self.queue = PriorityQueue()
        self.max_concurrency = max_concurrency
        self.engine_list = self.init_engines(
            max_concurrency, capability_info, workflow_service
        )
        self.queue_manager = threading.Thread(
            target=self.process_executions, daemon=True
        )
        self.queue_manager.start()
        self.paused = False

    @classmethod
    def init_engines(
            cls,
            num_engines: int,
            capability_info: CapabilityInfoIF,
            workflow_service: WorkflowServiceIF,
    ) -> NamedTuple:
        """
        Initialize available engine queue and list of in-use engines

        :return: EngineList
        """
        engine_list = cls.EngineList(Queue(), [])
        for _ in range(num_engines):
            engine_list.available.put(
                CapabilityEngine(capability_info, workflow_service)
            )

        return engine_list

    def process_executions(self):
        """
        Process execution at the front of the queue; I imagine this would be called repeatedly
        with a small delay
        """
        while True:
            if self.queue.empty() is False:
                execution = self.queue.get()[1]
                engine = self.get_available_engine()
                if engine:
                    execution = self.capability_info.lookup_execution(execution.id)
                    # Start engine
                    engine.execute(execution)
                    # Move engine to in-use list
                    self.engine_list.in_use.append(engine)
                else:
                    # FIXME: Logging
                    print("No available engines. Try again later.")
            time.sleep(0.5)

    def find_engine(self, execution_id: int) -> CapabilityEngine:
        """
        Find in-use engine executing execution with given ID

        :param execution_id: ID of requested execution
        :return: Corresponding engine
        """
        for engine in self.engine_list.in_use:
            if engine.execution_id == execution_id:
                # Found correct engine
                return engine

    def complete_execution(self, execution_id: int):
        """
        Remove execution from its engine and return engine to available queue

        :param execution_id: ID of requested execution
        """
        engine = self.find_engine(execution_id)
        self.engine_list.in_use.remove(engine)
        engine.execution_id = None
        engine.execution = None
        self.engine_list.available.put(engine)

    def get_available_engine(self) -> Optional[CapabilityEngine]:
        """
        Get an engine if one is available

        :return: Engine if one is available, else False
        """
        try:
            return self.engine_list.available.get_nowait()
        except Empty:
            return None

    def enqueue(self, request: CapabilityExecutionIF, priority: int):
        self.queue.put((priority, request))

    def dequeue(self) -> CapabilityExecutionIF:
        return self.queue.get()

    def pause(self):
        # TODO: Implement
        self.paused = True

    def unpause(self):
        # TODO: Implement
        self.paused = False

    def __len__(self):
        return self.queue.qsize()
