"""
Services defined by the Workspaces system, to be used by our APIs and client programs.
"""

import os
import stat
import subprocess
import threading
from pathlib import Path
from tempfile import mkdtemp
from typing import Dict, List, Union

import requests
from channels.amqp_helpers import Channel, WorkflowEventChannel
from pycapo import CapoConfig
from sqlalchemy.orm import Session

from workspaces.workflow_interfaces import WorkflowRequestIF
from workspaces.notification_interfaces import NotificationInfoIF
from .helpers import WorkflowRequestState
from .schema import (
    AbstractFile,
    Base,
    NotificationTemplate,
    Workflow,
    WorkflowEvent,
    WorkflowEventType,
    WorkflowRequest,
    WorkflowRequestFile,
    WorkflowTemplate,
)
from .workflow_interfaces import WorkflowIF, WorkflowInfoIF, WorkflowServiceIF


class WorkflowServiceRESTClient(WorkflowServiceIF):
    def __init__(self):
        self.url = (
            CapoConfig()
            .settings("edu.nrao.archive.workspaces.WorkflowSettings")
            .serviceUrl
        )

    def execute(self, request: WorkflowRequestIF, files: List["AbstractFile"]):
        # step 1: create the request
        req = requests.post(
            f"{self.url}/workflows/{request.workflow_name}/requests/create?args={request.argument}"
        ).json()

        # step 2: if necessary, pass the files up for this request
        for file in files:
            requests.post(
                f"{self.url}/workflows/requests/{req['workflow_request_id']}/files/{file.filename}",
                body=file.content,
            )

        # step 3: execute the request
        requests.post(
            f"{self.url}/workflows/requests/{req['workflow_request_id']}/submit"
        )


class WorkflowService(WorkflowServiceIF):
    """
    Executes workflows; should be a freestanding service.
    """

    # The next few things that need to happen here are in response to
    # WorkflowEvents that we receive from wf_monitor as the workflow
    # execution evolves. So we have to set up listening at some point
    # in this class
    def __init__(self, info: "WorkflowInfoIF"):
        self.info = info

        # Start listening for events from the wf_monitor stream
        self.listener = threading.Thread(
            target=lambda: Channel(WorkflowEventChannel(), threaded=True).listen(
                self.on_workflow_event
            )
        )
        self.listener.start()

    def execute(self, request: WorkflowRequest, files: List[AbstractFile] = None):
        """
        Execute a workflow per the supplied parameters.

        :param request: The Workflow Request to execute
        :param files: additional files to execute with request
        :return:
        """
        # 1. look up workflow, returns workflow
        definition = self.info.lookup_workflow_definition(request.workflow_name)

        # 2. render templates to files, returns list of rendered files
        workflow_files = definition.render_templates(request.argument, files)
        for file in workflow_files:
            self.info.save_file(
                request=request, filename=file.filename, content=file.content
            )

        # 4. prepare files for condor execution
        if not request.results_dir:
            # Create temp results directory if the request doesn't already have one (from a previous execution attempt)
            request.results_dir = str(
                self._prepare_files_for_condor(workflow_files).absolute()
            )
            self.info.save_request(request)

        # 5. execute condor and retrieve log file
        log_file = self._execute_prepared(Path(request.results_dir))

        # 6. start reading the log file (background it)
        subprocess.run(
            ["wf_monitor", log_file, str(request.workflow_request_id)],
            cwd=request.results_dir,
        )

    @staticmethod
    def _prepare_files_for_condor(files: List[WorkflowRequestFile]) -> Path:
        """
        Place the files for Condor into a new temp directory and returns the directory.

        :param files:  a dictionary of filename -> content
        :return:       a Path
        """
        # 1. create a temporary directory
        # TODO: decide where to put temp directory and make capo property
        if not os.path.isdir(str(Path.home()) + "/workspaces_tmp"):
            os.mkdir(str(Path.home()) + "/workspaces_tmp")
        temp_folder = Path(mkdtemp(dir=str(Path.home()) + "/workspaces_tmp"))

        # 2. spool each of the temp files to it
        for file in files:
            (temp_folder / file.filename).write_bytes(file.content)

        # 3. make any scripts in there executable
        for file in temp_folder.glob("*.sh"):
            file.chmod(file.stat().st_mode | stat.S_IEXEC)

        # finished, return folder
        return temp_folder

    @staticmethod
    def _execute_prepared(folder: Path) -> Path:
        """
        Execute HTCondor using the named folder as the source of the files.

        :param folder:  the path to the folder to execute
        :return:        the path to the log file
        """
        print(f"executing on folder {folder}")

        # some file in here should end in .dag; that file is our dagman input
        # TODO: not currently using DAG files for vulture
        dagman = list(folder.glob("*.dag"))[0]
        print(f"dagman file {dagman} exists.")

        # need .condor file for vulture submission
        condor = list(folder.glob("*.condor"))[0]
        print(f"condor file {condor} exists.")

        # ensure the log file exists
        logfile = folder / "condor.log"
        print(f"log file {logfile} exists.")
        logfile.touch()

        # submit
        # print("submitting to condor...")
        # subprocess.run(["condor_submit_dag", str(dagman)], cwd=str(folder.absolute()))

        # vulture is a workaround for testing locally without submitting to condor
        print("submitting to vulture...")
        subprocess.run(
            ["vulture", "job", "execute", str(condor)], cwd=str(folder.absolute())
        )

        # return the logfile
        return logfile

    def on_workflow_event(self, event: WorkflowEvent):
        request = self.info.lookup_workflow_request(event.workflow_request_id)

        status: str
        # 2. update request record with new status
        if event.type == WorkflowEventType.EXECUTING:
            status = WorkflowRequestState.Running.name
        elif event.type == WorkflowEventType.TERMINATED:
            if event.retval == 0:
                status = WorkflowRequestState.Complete.name
            else:
                status = WorkflowRequestState.Failed.name
        else:
            status = WorkflowRequestState.Running.name

        print(
            f"Updating state on workflow request {request.workflow_request_id} to {status}..."
        )
        request.update_status(status)


class WorkflowInfo(WorkflowInfoIF):
    """
    Implements WorkflowInfo, backed by a relational database, using SQL Alchemy.
    """

    def __init__(self, session):
        self.session = session

    def lookup_workflow_definition(self, name: str) -> Workflow:
        return self.session.query(Workflow).get(name)

    def lookup_workflow_request(self, request_id: int) -> WorkflowRequest:
        return self.session.query(WorkflowRequest).get(request_id)

    def lookup_workflow_templates_for(
        self, workflow_name: str
    ) -> List[WorkflowTemplate]:
        """
        Queries the workflow_templates table for all templates belonging to a given workflow

        :param workflow_name: Name of workflow to lookup templates for
        :return: List of templates
        """
        return (
            self.session.query(WorkflowTemplate)
            .filter_by(workflow_name=workflow_name)
            .all()
        )

    def all_workflows(self) -> List[Workflow]:
        return self.session.query(Workflow).all()

    def all_workflow_requests(self) -> List[WorkflowRequest]:
        """
        Query for the list of all workflow requests

        :return: List of workflow requests
        """
        return self.session.query(WorkflowRequest).all()

    def create_workflow_request(
        self, workflow: Union[str, WorkflowIF], argument: Dict
    ) -> WorkflowRequest:
        """
        Create new workflow request and save to database

        :param workflow: workflow to run
        :param argument: workflow arguments
        :return: new WorkflowRequest
        """

        # if it's an object with a workflow_name, then that's what we use,
        # otherwise, the argument must be already be a string
        workflow_name = (
            workflow.workflow_name if hasattr(workflow, "workflow_name") else workflow
        )

        request = WorkflowRequest(
            workflow_name=workflow_name,
            argument=argument,
            state=WorkflowRequestState.Created.name,
            results_dir="",
        )
        self.save_request(request)
        print(f"Created Workflow Request {request}")
        return request

    def save_request(self, request: WorkflowRequest):
        """
        Save a given request and return an integer identifier for it

        :param request: request to save
        """
        self.session.add(request)
        self.session.flush()

    def save_file(
        self, request: WorkflowRequest, filename: str, content: bytes
    ) -> WorkflowRequestFile:
        """
        Save a given file for the specified request to the database

        :param request: request to update with new file
        :param filename: filename of new file
        :param content: contents of new file in bytes
        :return:
        """
        wrf = WorkflowRequestFile(
            workflow_request_id=request.workflow_request_id,
            filename=filename,
            content=content,
        )
        self.session.add(wrf)
        self.session.flush()
        return wrf


class NotificationInfo(NotificationInfoIF):
    def __init__(self, session: Session):
        self.session = session

    def all_templates(self) -> List[NotificationTemplate]:
        return self.session.query(NotificationTemplate).all()

    def lookup_template(self, template_name: str) -> NotificationTemplate:
        return self.session.query(NotificationTemplate).filter_by(name=template_name).first()

    def create_template(
            self, name: str, description: str, template: str
    ) -> NotificationTemplate:
        """
        Create new notification template and save it in the database

        :param name: Name of new template
        :param description: brief description of template
        :param template: the full text of the template
        :return: Created NotificationTemplate
        """
        template = NotificationTemplate(name=name, description=description, content=template)
        self.save_entity(template)
        return template


    def delete_template(self, template_name:str):
        template = self.session.query(NotificationTemplate).filter_by(name=template_name).first()
        self.session.delete(template)


    def save_entity(self, entity: Base):
        """
        Save entity to database

        :param entity:
        """
        self.session.add(entity)
        self.session.flush()
