"""
These tests check test data for various types of products,
so we'll know we're getting what we expect when we use the test data in
our tests.
"""

import pytest

from shared.workspaces.test.test_data.utilities import Deliverable, DeliverableProduct

from .test_data.products.expected_values_alma import (
    AlmaAudiProductSet,
    AlmaProprietaryImageProductSet,
    AlmaPublicProductSet,
)
from .test_data.products.expected_values_evla import (
    PPR_FILENAME,
    CalibrationProduct,
    VlaProductSet,
    VlassImageProductSet,
    VlbaProductSet,
)
from .test_data.products.expected_values_gbt import GbtProductSet

MS_TGZ = ".ms.tgz"
WEBLOG = "weblog.tgz"

# pylint: disable=C0301, R0201, R0912


@pytest.mark.skip(
    reason="Test currently failing within Docker image; this is being looked at"
)
def test_gets_evla_eb_products():
    """
    Check our execution blocks test data
    :return:
    """

    product_set = VlaProductSet()
    exec_blocks = product_set.exec_blocks
    assert len(exec_blocks) == 2
    assert len(product_set.project.authors) == 5
    assert product_set.total_size == 95649344815
    assert len(product_set.science_products) == len(exec_blocks)


@pytest.mark.skip(
    reason="Test currently failing within Docker image; this is being looked at"
)
def test_gets_evla_ms():
    """
    Check our basic measurement set test data
    :return:
    """

    product_set = VlaProductSet()
    basic_ms = product_set.measurement_set
    expected_file_count = 9
    for filegroup in basic_ms:
        files = basic_ms[filegroup]
        actual_file_count = 0
        for item in files:
            actual_file_count += 1
            filename = str(item["filename"])

            # spot-check a few files
            if filename == "unknown.auxproducts.tgz":
                assert item["size"] == 615
            if filename.startswith("17A-109") and filename.endswith(MS_TGZ):
                if filegroup.endswith("18468"):
                    assert item["size"] == 23690977531
                elif filename.endswith("41979"):
                    assert item["size"] == 23690977245
            elif filename == "unknown.pprequest.xml":
                if filegroup.endswith("18468"):
                    assert item["size"] == 5610
                elif filename.endswith("41979"):
                    assert item["size"] == 5611
            elif filename == WEBLOG:
                if filegroup.endswith("18468"):
                    assert item["size"] == 1681803
                elif filename.endswith("41979"):
                    assert item["size"] == 1683135

        assert actual_file_count == expected_file_count


@pytest.mark.skip(
    reason="Test currently failing within Docker image; this is being looked at"
)
def test_gets_evla_cms():
    """
    Check our calibrated measurement set ('restore') test data
    :return:
    """

    product_set = VlaProductSet()
    cms = product_set.cms
    expected_file_count = 8
    for filegroup in cms:
        files = cms[filegroup]
        actual_file_count = 0
        for item in files:
            actual_file_count += 1
            filename = str(item["filename"])

            # spot-check a few files
            if filename == PPR_FILENAME:
                assert item["size"] == 1632
            if filename.startswith("17A-109") and filename.endswith(".tar.tgz"):
                if filegroup.endswith("18468"):
                    assert item["size"] == 47520134173
                elif filename.endswith("41979"):
                    assert item["size"] == 88842463132
            elif filename == "unknown.pipeline_manifest.xml":
                if filegroup.endswith("18468"):
                    assert item["size"] == 24025
                elif filename.endswith("41979"):
                    assert item["size"] == 11006
            elif filename == "casa_commands.log":
                if filegroup.endswith("18468"):
                    assert item["size"] == 260953
                elif filename.endswith("41979"):
                    assert item["size"] == 133935

        assert actual_file_count == expected_file_count


def test_gets_image_product():
    """Confirm one test image product with correct size and
    correct project info
    """

    product_set = VlassImageProductSet()
    assert product_set is not None
    assert not hasattr(product_set, "exec_blocks")
    assert product_set.project.project_code == "VLASS1.1"
    assert len(product_set.project.authors) == 5
    assert len(product_set.science_products) == 1
    assert product_set.total_size == 110851200


def test_gets_vlba_products():
    """Confirm that we get expected number of VLBA products
    with correct size and project info
    """

    product_set = VlbaProductSet()
    assert product_set is not None
    exec_blocks = product_set.exec_blocks
    assert len(exec_blocks) == 4
    assert len(product_set.project.authors) == 4
    assert product_set.total_size == 2140560000
    assert len(product_set.science_products) == len(exec_blocks)


def test_gets_calibration():
    """Confirm that we get one calibration product with correct size and
    correct project info
    """

    product = CalibrationProduct()
    assert product is not None
    assert not hasattr(product, "exec_blocks")
    assert len(product.project.authors) == 9
    science_product = product.science_product
    assert science_product.science_product_type == "Calibration"
    assert science_product.external_name == "18B-265_2019_12_10_T00_00_59.203.tar"
    assert len(product.file_info) == 1
    assert product.total_size == 27668480


def test_gets_alma_raw_data():
    """ Grab ALMA SDM+BDF test data and confirm we got what we expect """

    product_set = AlmaPublicProductSet()

    file_info = product_set.file_info
    assert isinstance(file_info, dict)
    assert len(file_info) == 3
    assert product_set.rawdata_total_size == 1368988450

    file_sizes = dict()
    delivered_products = list()
    for file_type_str in file_info.keys():
        values = file_info[file_type_str]
        file_type = Deliverable.from_str(file_type_str)
        assert file_type is not None
        file_sizes[file_type] = file_info.values()
        delivered_products.append(DeliverableProduct(file_type, values))

    # spot-check what we got for each kind of deliverable
    for delivery in delivered_products:
        file_count = len(delivery.file_info)
        if delivery.type == Deliverable.TAR:
            assert file_count == 1
            for filename, size in delivery.file_info.items():
                assert filename.startswith(
                    "uid___A002_Xcd8029_Xb0a4"
                ) and filename.endswith(".tar.gz")
                assert size == 457195765
        elif delivery.type == Deliverable.SDM:
            file_count = len(delivery.file_info)
            assert file_count == 39
            xml_count = 0
            bin_count = 0
            for filename, size in delivery.file_info.items():
                if filename.endswith(".xml"):
                    xml_count += 1
                    if filename.startswith("Pointing."):
                        assert size == 963
                elif filename.endswith("bin"):
                    bin_count += 1
                    if filename.startswith("Pointing."):
                        assert size == 92853427

            assert xml_count == 36
            assert bin_count == 3
        elif delivery.type == Deliverable.BDF:
            file_count = len(delivery.file_info)
            assert file_count == 216
            for file_path, size in delivery.file_info.items():
                assert file_path.startswith("ASDMBinary")
                assert size >= 14000
        else:
            pytest.fail(f"Unexpected deliverable from ALMA raw data: {delivery.type}")


def test_gets_alma_basic_ms():
    """
    Analyze our ALMA basic MS test data and ensure the values are as we expect

    """

    product_set = AlmaPublicProductSet()
    basic_ms = product_set.measurement_set
    assert isinstance(basic_ms, dict)
    assert len(basic_ms) == 3

    for name in basic_ms.keys():
        file_info = basic_ms[name]
        assert isinstance(file_info, dict)
        assert len(file_info) == 10
        weblog_found = False
        ms_tgz_found = False
        for filename, size in file_info.items():
            if filename in (PPR_FILENAME, "unknown.download_alma_ms.pprequest.xml"):
                assert size in (4582, 4584)
            # spot-check some file sizes
            if name == "uid___A002_Xcd8029_Xa892":
                if filename == WEBLOG:
                    assert size == 1592335
                    weblog_found = True
                elif filename.endswith(MS_TGZ):
                    assert size == 519234359
                    ms_tgz_found = True
            elif name == "uid___A002_Xcd8029_Xb0a4":
                if filename == "weblog.tgz":
                    assert size == 1576475
                    weblog_found = True
                elif filename.endswith(MS_TGZ):
                    assert size == 518582184
                    ms_tgz_found = True
            elif name == "uid___A002_Xcd8029_Xfdd":
                if filename == WEBLOG:
                    assert size == 1562579
                    weblog_found = True
                elif filename.endswith(MS_TGZ):
                    assert size == 504128830
                    ms_tgz_found = True
            else:
                pytest.fail(f"Unexpected item found in basic MS: {name}")
        assert weblog_found and ms_tgz_found


def test_gets_alma_restored_ms():
    """We're expecting these files inside the tar returned
    for ALMA restored MS download of 2017.1.00297.S
    """
    product_set = AlmaPublicProductSet()
    restored_ms = product_set.restored_ms
    assert isinstance(restored_ms, dict)
    assert len(restored_ms) == 1
    assert product_set.rest_ms_size == 3947540021

    for name in restored_ms.keys():
        assert name == "uid___A002_Xcd8029_Xb0a4"
        values = restored_ms[name]
        assert len(values) == 4
        assert values["uid___A002_Xcd8029_Xb0a4.ms.tgz"] == 1325422495
        assert values["uid___A002_Xcd8029_Xa892.ms.tgz"] == 1331336708
        assert values["uid___A002_Xcd8029_Xfdd.ms.tgz"] == 1290775660
        assert values[PPR_FILENAME] == 5158


def test_gets_alma_images():
    """Expected download results for a proprietary ALMA project
    with image data
    """

    product = AlmaProprietaryImageProductSet()
    file_info = product.file_info
    assert len(file_info) == 3
    total_size = 0
    for filename, size in file_info.items():
        total_size += size
        if "_71595054" in filename or "_4d1b66da" in filename:
            assert size == 161280
        else:
            assert "_image_cube_daf0d76e" in filename
            assert size == 112907520

    assert total_size == product.total_size


def test_gets_audi_product():
    """
    Make sure we're getting the product and files we expect for our
    AlmaAudiProductSet

    :return:
    """

    product = AlmaAudiProductSet()
    assert product.total_size == 1368988450
    file_info = product.file_info
    assert len(file_info) == 3
    for file_type in ["SDM", "BDF", "TAR"]:
        assert file_type in file_info.keys()


def test_gets_gbt_product():
    """ make sure we're getting the GBT test data values we expect """
    product_set = GbtProductSet()
    assert product_set.project.project_code == "AGBT17B_044"
    assert product_set.total_size == 948572979200
    assert len(product_set.file_info) == 4
